@@ -1,6 +1,6 @@
 #!/usr/bin/perl
 package SVN::Mirror;
-our $VERSION = '0.51';
+our $VERSION = '0.61';
 use SVN::Core;
 use SVN::Repos;
 use SVN::Fs;
@@ -14,7 +14,7 @@ SVN::Mirror - Mirror remote repository to local Subversion repository
 =head1 SYNOPSIS
 
  my $m = SVN::Mirror->new (source => $url,
-			   repost => '/path/to/repository',
+			   repos => '/path/to/repository',
 			   target_path => '/mirror/project1'
 			   repos_create => 1,
 			   skip_to => 100
@@ -51,9 +51,12 @@ sub _schema_class {
     my ($url) = @_;
     die "no source specificed" unless $url;
     return 'SVN::Mirror::Ra' if $url =~ m/^(https?|file|svn(\+.*?)?):/;
-    return 'SVN::Mirror::VCP' if $url =~ m/^(p4|cvs|arch)/ and eval {
-	require SVN::Mirror::VCP; 1
-    };
+    if ($url =~ m/^(p4|cvs|arch)/) {
+	eval {
+	    require SVN::Mirror::VCP; 1
+	} and return 'SVN::Mirror::VCP';
+	warn "VCP required.  Please install VCP and VCP::Dest::svk.\n";
+    }
 
     die "schema for $url not handled\n";
 }
@@ -115,7 +118,6 @@ sub has_local {
     return unless $specanchor;
     my $path = $mirrored{$specanchor};
     $spec =~ s/^\Q$specanchor\E//;
-    $path =~ s/^\Q$spec\E//;
     my $m = SVN::Mirror->new (target_path => $path,
 			     repos => $repos,
 			     pool => SVN::Pool->new,
@@ -157,8 +159,12 @@ sub is_mirrored {
 
 sub load_fromrev {
     my ($self) = @_;
-    my $changed = $self->{root}->node_created_rev ($self->{target_path});
-    my $prop = $self->{fs}->revision_prop ($changed, 'svm:headrev');
+    $self->lock('mirror');
+    my $fs = $self->{fs};
+    my $root = $fs->revision_root ($fs->youngest_rev);
+    my $changed = $root->node_created_rev ($self->{target_path});
+    $self->unlock('mirror');
+    my $prop = $fs->revision_prop ($changed, 'svm:headrev');
     return unless $prop;
     my %revs = map {split (':', $_)} $prop =~ m/^.*$/mg;
     my $uuid = $self->{rsource_uuid} || $self->{source_uuid};
@@ -170,11 +176,16 @@ my %CACHE;
 
 sub find_local_rev {
     my ($self, $rrev, $uuid) = @_;
-    return if $rrev > ($self->{working} || $self->{fromrev});
-    my $pool = SVN::Pool->new_default ($self->{pool});
 
-    my $fs = $self->{repos}->fs;
+    # if uuid is the repository we talk to directly, return
+    # null for revisions larger than what we have
     $uuid ||= $self->{source_uuid};
+    return if $uuid eq $self->{rsource_uuid} &&
+	$rrev > ($self->{working} || $self->{fromrev});
+
+    my $pool = SVN::Pool->new_default ($self->{pool});
+    my $fs = $self->{repos}->fs;
+
     # XXX: make better use of the cache
     my $cache = $CACHE{$fs->get_uuid}{$uuid} ||= {};
     return $cache->{$rrev} if exists $cache->{$rrev};
@@ -196,7 +207,8 @@ sub find_local_rev {
         $old_pool->clear;
         ($old_pool, $new_pool) = ($new_pool, $old_pool);
 	my ($lrev) = $prop =~ m/^\Q$uuid\E:(\d+)$/m;
-        next unless defined $lrev;
+	# 0 would be the init change we had. not good for any use.
+        next unless $lrev;
 	$cache->{$lrev} = $rev;
 	# XXX: make use of svm:incomplete prop to see if this is safe to use
 	return $rev if $rrev >= $lrev;
@@ -237,6 +249,7 @@ sub delete {
         $txnroot->change_node_prop ($self->{target_path}, 'svm:uuid', undef);
     }
     my $rev = $self->commit_txn($txn);
+    delete $CACHE{$fs->get_uuid}{$self->{source_uuid}};
     print "Committed revision $rev.\n";
 }
 
@@ -311,7 +324,8 @@ sub relocate {
     $self->load_state ();
 
     my $ra = $self->_new_ra (url => $self->{source});
-    die "uuid is different" unless $ra->get_uuid eq $self->{source_uuid};
+    my $ra_uuid = $ra->get_uuid;
+    die "Local and remote UUID differ." unless ($ra_uuid eq $self->{source_uuid} or $ra_uuid eq $self->{rsource_uuid});
 
     # Get latest revprops
     my $old_prevs = $self->{fs}->revision_proplist(
@@ -356,7 +370,7 @@ sub mkpdir {
 	$path = File::Spec::Unix->join($path, shift @dirs);
 	my $kind = $self->{root}->check_path ($path);
 	if ($kind == $SVN::Core::node_none) {
-	    $root->make_dir ($path, $self->{pool});
+	    $root->make_dir ($path, SVN::Pool->new);
 	    $new = 1;
 	}
 	elsif ($kind != $SVN::Core::node_dir) {
@@ -420,18 +434,63 @@ sub upgrade {
 
 sub commit_txn {
     my ($self, $txn) = @_;
-    my $rev;
+    return ($txn->commit)[1];
+}
 
-    if ($^O eq 'MSWin32' and ref($self) and -e "$self->{repospath}/db/current") {
-        # XXX - On Win32+fsfs, ->commit works but raises exceptions
-        eval { $txn->commit };
-        $rev = $self->{fs}->youngest_rev;
+use Sys::Hostname;
+
+sub _lock_token {
+    my $token = $_[0]->{target_path};
+    $token =~ s/_/__/g;
+    $token =~ s{/}{_}g;
+    return "svm:lock:$_[1]:$token";
+}
+
+sub lock {
+    my ($self, $what) = @_;
+    my $fs = $self->{fs};
+    my $token = $self->_lock_token ($what);
+    my $content = hostname.':'.$$;
+    my $where = join(' ', (caller(0))[0..2]);
+    die $where."\n".$self->{locked}{$what} if exists $self->{locked}{$what};
+    # This is not good enough but race condition should result in failed sync
+    # without corrupting repository.
+    LOCKED:
+    {
+	while (1) {
+	    my $who = $fs->revision_prop (0, $token) or last LOCKED;
+	    if ($who eq $content) {
+		$self->unlock ($what);
+		Carp::confess "recursive lock? $what $where $self->{locked}{$what}";
+	    }
+	    if ($self->{lock_message}) {
+		$self->{lock_message}->($self, $what, $who);
+	    }
+	    else {
+		print "Waiting for $what lock on $self->{target_path}: $who.\n";
+	    }
+	    sleep 1;
+	}
     }
-    else {
-        $rev = ($txn->commit)[1];
+    $fs->change_rev_prop (0, $token, $content);
+    $self->{locked}{$what} = $where;
+}
+
+sub unlock {
+    my ($self, $what) = @_;
+    if ($what eq 'force') {
+	for (keys %{$self->{fs}->revision_proplist(0)}) {
+	    $self->{fs}->change_rev_prop (0, $_, undef);
+	}
+	delete $self->{locked};
+	return;
     }
 
-    return $rev;
+    my $token = $self->_lock_token ($what);
+    if ($self->{locked}{$what}) {
+	$self->{fs}->change_rev_prop (0, $token, undef);
+	delete $self->{locked}{$what};
+    }
 }
 
 =head1 AUTHORS
@@ -440,7 +499,7 @@ Chia-liang Kao E<lt>clkao@clkao.orgE<gt>
 
 =head1 COPYRIGHT
 
-Copyright 2003-2004 by Chia-liang Kao E<lt>clkao@clkao.orgE<gt>.
+Copyright 2003-2005 by Chia-liang Kao E<lt>clkao@clkao.orgE<gt>.
 
 This program is free software; you can redistribute it and/or modify it
 under the same terms as Perl itself.
